<?php

namespace JimsCustomerBooking\Database;

use JimsCustomerBooking\Helpers\StringHelper;
use JimsCustomerBooking\JimsPlugin;

/**
 * Class Database.
 */
class Database
{
    public static $queries_to_run = [];

    /**
     * List of migrations to run
     * a migration is a function that creates or modifies a table
     * the migrations array has the version as the key, and the name of the migration
     * function as the value.
     *
     * Every migration function name should start with 'migration_'
     */
    public static array $migrations = [
        '3.1.03' => 'migrationAddErrorCodeColumn',
        '3.3.01' => 'migrationAddDivisionIDColumn',
        '3.3.03' => 'migrationAddModeAndChannelColumns',
    ];

    public static function setup(): void
    {
        if (Database::tableExists('customer_bookings')) {
            return;
        }

        self::createTable([
            'id mediumint(9) NOT NULL AUTO_INCREMENT',
            'booking_id varchar(255) NOT NULL',
            'division_id int(255) NULL',
            'title varchar(255) NULL',
            'data longtext NOT NULL',
            'status varchar(255) NOT NULL',
            'error_code varchar(255) NULL',
            'mode varchar(255) NULL',
            'channel varchar(255) NULL',
            'created_at datetime DEFAULT \'0000-00-00 00:00:00\' NOT NULL',
            'updated_at datetime DEFAULT \'0000-00-00 00:00:00\' NOT NULL',
            'PRIMARY KEY (id)',
        ], 'customer_bookings');
    }

    public static function setupOrUpdateDatabase($current_version = '0.0.0')
    {
        error_log('setupOrUpdateDatabase', 0);
        self::setup();

        self::update($current_version);

        $result = self::run();
    }

    public static function run()
    {
        return self::runSQL(self::$queries_to_run, true);
    }

    /**
     * @param mixed $oldVersion
     */
    public static function update($oldVersion): array
    {
        $output = [];

        if ($oldVersion === '0.0.0') {
            $oldVersion = '3.1.0'; // set the first version with migration system
        }

        // run all the migration functions that are greater than the old version
        foreach (self::$migrations as $version => $migration) {
            if (StringHelper::compareVersionStrings($version, $oldVersion)) {
                array_push($output, [
                    'version' => $version,
                    'result' => call_user_func_array(self::class . '::' . $migration, []),
                ]);
            }
        }

        return $output;
    }

    /**
     * Creates a table in the database, from an array of columns.
     *
     * @param mixed $table_name
     */
    public static function createTable(array $columns, $table_name): bool
    {
        $bookings_table_name = '';
        $sql = self::getCreateTableSQL($columns, $table_name, $bookings_table_name);

        self::addQueryToRun($sql);

        return true;
    }

    /**
     * @param mixed $table_name
     * @param mixed $bookings_table_name
     *
     * @return false|string
     */
    public static function getCreateTableSQL(array $columns, $table_name, &$bookings_table_name)
    {
        global $wpdb;

        if (count($columns) == 0) {
            return false;
        }

        $charset_collate = $wpdb->get_charset_collate();

        $bookings_table_name = self::getTableName($table_name);

        $columns = implode(',', $columns);

        return "CREATE TABLE IF NOT EXISTS {$bookings_table_name} ({$columns}) {$charset_collate}";
    }

    /**
     * @return null|bool|int|\mysqli_result|resource
     */
    public static function deleteAllTransients()
    {
        global $wpdb;

        $table_name = $wpdb->prefix . 'options';
        $plugin_prefix = JimsPlugin::PREFIX;

        // phpcs:ignore
        $sql = "DELETE FROM {$table_name} WHERE `option_name` like '%transient%' and `option_name` like '%{$plugin_prefix}%';";

        return $wpdb->query($sql);
    }

    /**
     * @param mixed $queries
     */
    public static function runSQL($queries, bool $use_dbDelta = true): array
    {
        $result = [];

        if ($use_dbDelta) {
            require_once ABSPATH . 'wp-admin/includes/upgrade.php';
            $result = dbDelta($queries);

            if (!is_array($result)) {
                $result = [$result];
            }
        }

        // import the global wpdb object
        global $wpdb;

        if (count($result) < count($queries) + 1) {
            $result = array_map(function ($query) use ($wpdb) {
                return $wpdb->query($query);
            }, $queries);
        }

        return $result;
    }

    public static function deleteDatabase(): bool
    {
        global $wpdb;

        $table_name = self::getTableName();

        $sql = "DROP TABLE IF EXISTS {$table_name}";

        return $wpdb->query($sql);
    }

    public static function addQueryToRun(string $query): void
    {
        self::$queries_to_run[] = $query;
    }

    public static function tableExists($table_name): bool
    {
        global $wpdb;

        $table_name = self::getTableName($table_name);

        $sql = "SHOW TABLES LIKE '{$table_name}'";

        return $wpdb->get_var($sql) == $table_name;
    }

    public static function dbLog($message): void
    {
        error_log(json_encode($message), 0, '/var/www/jims-wp-default.local/wp-content/plugins/jims-wpcb/db.log');
    }

    // Migration Functions

    public static function migrationAddModeAndChannelColumns(): string
    {
        $table_name = self::getTableName('customer_bookings');

        // add mode and channel columns after status column
        // phpcs:ignore
        $sql = "ALTER TABLE {$table_name} ADD COLUMN mode varchar(255) NULL AFTER error_code, ADD COLUMN channel varchar(255) NULL AFTER mode";

        self::addQueryToRun($sql);

        return $sql;
    }

    public static function migrationAddErrorCodeColumn(): string
    {
        $table_name = self::getTableName('customer_bookings');

        // add error_code column after status column
        $sql = "ALTER TABLE {$table_name} ADD COLUMN error_code varchar(255) NULL AFTER status";

        self::addQueryToRun($sql);

        return $sql;
    }

    public static function migrationAddDivisionIDColumn(): string
    {
        $table_name = self::getTableName('customer_bookings');

        // add error_code column after status column
        $sql = "ALTER TABLE {$table_name} ADD COLUMN division_id int(255) NULL AFTER booking_id";

        self::addQueryToRun($sql);

        return $sql;
    }

    /**
     * @param mixed $name
     */
    protected static function getTableName($name = 'customer_bookings'): string
    {
        global $wpdb;

        return implode('_', [$wpdb->prefix . JimsPlugin::PREFIX, $name]);
    }
}
